
%% Glottal wave length computation using crosscorrelation teq
% Ales Zita 2011

%% Description
% Programm finds the extremal point of glottal waves 
% using Iterrative Normalized Cross Correlation technique
% Given image is pre-processed to enhance contrast and supress noise
% Such an image is than divided to parts, one for each chord opening
% using the data obtained as output from previous work of Mgr. David Hauzar 
% From these data first correlation kernel is computed
% Afeter cross correlating htis kernel with image, most extreme point is 
% found on the thresholded result of the correlation. This possition is used 
% as a base possition for next kernel and for glottal wave angle
% computation.
% Whole process is than repeted until it y-axis converges.
% Resultin possition is taken as local glottal wave maximum.
% Estimated glottal wave amplitude is computed as median of all local
% maxima.
% Estimated glottal wave angle is computed as mean value of all found local
% angles.

% test images : 
% Healthy & good quality imgs:  Martin, Hana, Hana2, Hana3, Hana4, Katya
% Unhealthy imgs :              VKG-014, VKG-019, VKG-028, VKG-030, VKG-031
% Bad resolution imgs :         40, 6


%% Code
% th = correlation threshold
% masksize = maximum half size of the mask (upper limit)
% mode : output mean 
%   - nothing(numeric values)
%   - 'display' - dispaly found wave maxima
%   - 'debug' - for debug purposes only !!!

% Example : GWProcess(0.5,8,'display')


function [LWAmp RWAmp LWAng RWAng] = GWProcess(imName,th,masksize,mode)

%% Parameters
% input parameters
if nargin == 0
    'Please input image name as function parameter'
end
if nargin == 1
    th = 0.5
    masksize = 8
end
if nargin < 4
    display = false
    debug = false
else
    if strcmp(mode, 'display')
        display = true;
        debug = false;
    else if strcmp(mode, 'debug')
            display = false;
            debug = true;
        else
            'Not supported mode'
        end
    end
end

% Built in program constants 
limPar = 6;
searchIterration = 5;
limit = max(limPar*(1-th));

% Image load and pre-processing; Image related data load & prep
J = imread(['Images/' imName '.jpg']);
%I = adapthisteq(J(:,:,1));  % adaptive histogram equalization
h = fspecial('gaussian');    % filter defintion
I = imfilter(J(:,:,1),h);          % filtering the image 
I = adapthisteq(I);
I = medfilt2(I);

opening = load (['Data/' imName '.openings.mat']);
a = opening.openingStartsHorizontalCoords;
b = opening.openingStartsVerticalCoords;
if (b(1) == 1) 
    b = b(2:size(b,2));
    a = a(1:size(b,2));
else
    b = b(1:size(a,2));
end
openings =[a;b]';

left = load(['Data/' imName '.leftopens.mat']);
leftopens = [left.temp, left.start'];
right = load(['Data/' imName '.rightopens.mat']);
rightopens = [right.temp, right.start'];

% Global variable deffinitions
LeftExtremaCoordinate = zeros(1,size(b,2)-1);
LeftAngleInstances = zeros(1,size(b,2)-1);
RightExtremaCoordinate = zeros(1,size(b,2)-1);
RightAngleInstances = zeros(1,size(b,2)-1);


%% Visualization & debug
if (display)
    mainFig = figure;imshow(J);text(-10,-10,imName);
    hold on;
end
if (debug)
    mainFig = figure;imshow(I);text(-10,-10,imName);
    hold on;
    plot(a,b,'o');
    plot(leftopens(:,1), leftopens(:,2), 'x');
    plot(rightopens(:,1), rightopens(:,2), 'x');
end

%% Main program body

% Loop 
for i=1:size(b,2)-1
    
    % Common data
    occurance = i;
    start = openings(occurance,:);
    leftamp = leftopens(occurance,:);
    rightamp = rightopens(occurance,:);
    
    %% Right vocal chord
    
    % Correlation kernel dimension computation 
    dimx = min(masksize,(start(1) - leftamp(1))/2);
    dimy = min(masksize,(leftamp(2) - start(2))/2);
    dim = floor(dimx+dimy);
    if mod(dim,2) == 0
        dim = dim - 1; %odd
    end
    
    % Direction of the glottal wave
    startangle = ComputeAngle(start,leftamp);
    angle = 2*pi -startangle;
    mask = CreateMask(dim, 'angled',angle);
    template = I(floor(leftamp(2)-dim/2+1):floor(leftamp(2)+dim/2), leftamp(1)-dim+1: leftamp(1), 1);
    template = uint8(mask).*template;
    %mask = MaskFromImage(template);
    %template = uint8(mask).*template;
    %template = uint8(mask).*template; % orez
    
    % Run Cross Correlation on cropped image
    part = I(max(b(i)-dim, 1):min(b(i+1)+dimy, size(I,1)),:,1); 
    cor = CrossCorrelation(double(template)/255,double(part)/255,double(mask)); 
    %ccc=uint8(((cor>th).*cor) * 255);
    
    % Compute the left extrema of correlation map
    startSearchx = floor(leftamp(1) - dim/2 +1);
    startSearchy = floor(leftamp(2)-b(i)+dim);
    [ex,ey] = FindLeftExtrema(cor>th,startSearchx,startSearchy); 
    
    % Re-evaluate the Wave direction 
    angle = ComputeAngle([startSearchx startSearchy],[ex ey]);
    RightWaveAngle = ComputeAngle(start,[ex ey+b(i)-dim]);    % Global wave directio aprox
    
    [ex,ey] = CorrelationWaveTracking(part, ex, ey, RightWaveAngle);
      
    % Result
    RightExtremaCoordinate(i) = ex;
    RightAngleInstances(i)    = RightWaveAngle;
    
    %% Debug
    if debug
        ey = ey + b(i) - dim;
        figure(mainFig), plot(ex, ey, 'x', 'Color', 'Red');
        plot(startSearchx,startSearchy+b(i)-dim,'x');
    end
    
    
    %% Left vocal chord
    
    % Correlation kernel dimension computation 
    dimx = min(masksize,(rightamp(1) - start(1))/2);
    dimy = min(masksize,(rightamp(2) - start(2))/2);
    dim = floor(dimx+dimy);
    if mod(dim,2) == 0
        dim = dim - 1; %odd
    end
    
    % Direction of the glottal wave
    startangle = ComputeAngle(start,rightamp);
    angle = pi/2 -startangle;
    mask = CreateMask(dim, 'angled',angle);
    template = I(floor(rightamp(2)-dim/2+1):floor(rightamp(2)+dim/2), rightamp(1)+1: rightamp(1)+dim,1);
    template = uint8(mask).*template;
    %mask = MaskFromImage(template);
    %template = uint8(mask).*template;
    
    %template = uint8(mask).*template; % orez
    
    % Run Cross Correlation on cropped image
    part = I(max(b(i)-dim, 1):min(b(i+1)+dimy, size(I,1)),:,1); % aby ne crosscorr na celym
    cor = CrossCorrelation(double(template)/255,double(part)/255,double(mask)); %maskovana CC
    %ccc=uint8(((cor>th).*cor) * 255);
    
    % Compute the right extrema of correlation map
    startSearchx = floor(rightamp(1) + dim/2 );
    startSearchy = floor(rightamp(2)-b(i)+dim);
    [ex,ey] = FindRightExtrema(cor>th,startSearchx,startSearchy); 
    
    % Re-evaluate the Wave direction 
    angle = ComputeAngle([startSearchx startSearchy],[ex ey]);
    LeftWaveAngle = ComputeAngle(start,[ex ey+b(i)-dim]);   % Global wave direction aprox
    
    [ex,ey] = CorrelationWaveTracking(part, ex, ey, LeftWaveAngle);
    
    % Result
    LeftExtremaCoordinate(i) = ex;
    LeftAngleInstances(i)    = LeftWaveAngle;
    
    %% Debug
    if debug
        ey = ey + b(i) - dim;
        figure(mainFig), plot(ex, ey, 'x','Color', 'Red');
        plot(startSearchx,startSearchy+b(i)-dim,'x');
    end
end

%% Result
%Left Wave Amplitude
sortCoord = sort(LeftExtremaCoordinate);
LWAmp = sortCoord(floor(size(sortCoord,2)/2));

%Left Wave Angle
LWAng = sum(sum(LeftAngleInstances))/numel(LeftAngleInstances);

%Right Wave Amplitude
sortCoord = sort(RightExtremaCoordinate);
RWAmp = sortCoord(floor(size(sortCoord,2)/2));

%Right Wave Angle
RWAng = sum(sum(RightAngleInstances))/numel(RightAngleInstances);


%% Visualization
if(display)
    figure(mainFig); 
    line([LWAmp LWAmp],[1 size(I,1)]);
    line([RWAmp RWAmp],[1 size(I,1)]);
end


%% Internal functions declarations

    function angle = ComputeAngle(startCoord, endCoord)
        if startCoord == endCoord
            angle = 0;
            return;
        end
        
        vectamp = endCoord - startCoord;
        angle = atan(vectamp(1)/vectamp(2));
        if (angle < 0), angle = angle + 2*pi; end;
    end

    function [ex,ey] = CorrelationWaveTracking(part, startx, starty, angle)
        
        % Init wave tracking variables
        searchIterration = 5;
        halfdim = floor(dim/2);
        ex  = startx;
        ey  = starty;
        
        prev = 0;
        count = 1;
        
        if (debug)
            test = figure; imshow(part);
            hold on;
            
            % Computing Step Visualization
            %figure(mainFig), line([start(1) startSearchx], [start(2) startSearchy+b(i)-dim]);
            %figure(mainFig), line([startSearchx ex], [startSearchy+b(i)-dim ey]);
            %figure(mainFig), plot(ex,ey,'x');
            
            
            % Angle Visualization
            %figure(mainFig), line([start(1) ex], [start(2) ey]);
            
            % Computing Step Visualization
            %figure(mainFig), line([start(1) startSearchx], [start(2) startSearchy+b(i)-dim]);
            %figure(mainFig), line([startSearchx ex], [startSearchy+b(i)-dim ey]);
            %figure(mainFig), plot(ex,ey,'x');
            
            % Angle Visualization
            %figure(mainFig), line([start(1) ex], [start(2) ey]);
        end
        
        while abs(prev - ex) > limit && count < searchIterration
            prev = ex;
            
            % Kernel boundaries
            tminx = max(1,ex-halfdim);
            tmaxx = min(size(part,2), ex+halfdim);
            tminy = max(1,ey-halfdim);
            tmaxy = min(size(part,1),ey+halfdim);
            
            % Kernel assign
            template = part(tminy:tmaxy, tminx:tmaxx);
            if size(template) ~= size(mask)
                d = min(min(size(template),size(mask)));
                template = template(1:d,1:d);
                mask = mask(1:d,1:d);
            end
            template = template.*uint8(mask);
            
            % Correlation
            cor = CrossCorrelation(double(template)/255,double(part)/255,double(mask));
            
            % Find new extrema
            [exNew,eyNew] = FindExtrema(cor>th,ex,ey,angle);
            
            % Re-evaluate the wave direction
            angle = ComputeAngle([ex ey], [exNew eyNew]);
            
            % Debug mode
            if (debug)
                % Computing Step Visualization
                imshow((((cor>th).*cor) * 255));
                figure(test), line([ex exNew], [ey eyNew]);
                figure(test), plot(ex,ey,'x');
                figure, imshow(cor);
                figure, imshow(template);
            end
            
            ex = exNew;
            ey = eyNew;
            
            count = 1 + count;
        end
        
        
    end


end

